import re
import pathlib
import numpy as np
from scipy.io import loadmat
from matplotlib.pyplot import imread

class logviewer:
    def __init__(self,
                 exp_dir : pathlib.Path,
                 logger = print):
        self.exp_dir = pathlib.Path(exp_dir)
        self.logger = logger
        if(self.exp_dir.is_dir()):
            self.logger('Looking for a log in: '+ str(self.exp_dir))
        else:
            self.logger('No such directory: ' + str(self.exp_dir))
        
    def get_variable(self, var_name, single_shot_index = -1, 
                     suffix = '.np*', mat_file_field = None):
        var_flist = list(self.exp_dir.glob(f'{var_name}*{suffix}'))
        if(len(var_flist) > 0):
            var_flist.sort()
            var_path = var_flist[single_shot_index]
            if(not var_path.is_file()):
                return
            self.logger(f'Loading {var_path}')
            if(var_path.suffix == '.npz'):
                buf = np.load(var_path)
                time_array = buf['time_array']
                n_logs = (time_array > 0).sum()
                time_array = time_array[:n_logs]
                data_array = buf['data_array']
                data_array = data_array[:n_logs]
                return(time_array, data_array)
            elif(var_path.suffix == '.npy'):
                return(np.load(var_path))
            elif(var_path.suffix == '.mat'):
                    if(mat_file_field is not None):
                        return loadmat(var_path)[mat_file_field]
                    else:
                        self.logger('You need to provide the field name for the mat file')
        else:
            var_dir = self.exp_dir / var_name
            if(var_dir.is_dir()):
                flist = list(var_dir.glob('*.*'))
                flist.sort()
                var_path = flist[single_shot_index]
                if(var_path.suffix == '.npy'):
                    return np.load(var_path)
                elif(var_path.suffix == '.mat'):
                    if(mat_file_field is not None):
                        return loadmat(var_path)[mat_file_field]
                    else:
                        self.logger('You need to provide the field name for the mat file')
    
    def get_images_as_stack(self, var_name, n_images : int = None):
        var_dir = self.exp_dir / var_name
        var_flist = list(var_dir.glob(f'*.*'))
        #img__5.302307367324829#
        if(len(var_flist) > 0):
            img_inds = np.zeros(len(var_flist))
            for fcnt, fpath in enumerate(var_flist):
                self.logger(fpath.stem)
                ints_in_stem = (re.findall('\d+', fpath.stem ))
                str1 = str(ints_in_stem[-2]) + '.' + str(ints_in_stem[-1])
                img_inds[fcnt] = float(str1)
            inds = np.argsort(img_inds)
            var_flist = [var_flist[ind] for ind in inds]
            if(n_images is None):
                n_images = len(var_flist)
            else:
                n_images = np.minimum(n_images, len(var_flist))
            var_flist = var_flist[:n_images]
            self.logger(f'There are {n_images} images to stack')
            img = imread(var_flist[0])
            img_set = np.zeros((n_images, ) + img.shape, dtype=img.dtype)
            self.logger(f'shape is: {img_set.shape}')
            for fcnt, fpath in enumerate(var_flist):
                img_set[fcnt] = imread(fpath)
            self.logger('Image stack is ready')
            return(img_set)
    
    def get_log_text(self, log_name='main_log'):
        flist = list(self.exp_dir.glob(f'{log_name}*.txt'))
        flist.sort()
        n_files = len(flist)
        if (n_files>0):
            txt = []
            for fcnt in range(n_files):
                with open(flist[fcnt]) as f_txt:
                    txt.append(f_txt.readlines())
            return txt


#!/usr/bin/env python

"""Tests for `lognflow` package."""

import pytest

from lognflow import lognflow, logviewer, printprogress

import numpy as np
import matplotlib.pyplot as plt

temp_dir = 'c:\Alireza\logs'

@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')

def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string

def test_lognflow():
    logger = lognflow(temp_dir)
    logger('This is a test for lognflow and log_var')    
    for _ in range(10000):
        logger(f'Log{_}'*200)

    logger('This is a new log file for another script', log_name = 'not_main_script')
    logger('For other log files you need to mention the log_name', log_name = 'not_main_script')

def test_log_var():
    logger = lognflow(temp_dir)
    logger('This is a test for lognflow and log_var')    

    for _ in range(1000):
        logger.log_var('vars/vec/v', np.random.rand(10000))
        
def test_log_animation():
    var1 = np.random.rand(32, 100, 100)
    logger = lognflow(temp_dir)
    logger('This is a test for log_animation')    
    
    logger.log_animation('var1',var1)

def test_log_single():
    var1 = np.random.rand(100)
    
    logger = lognflow(temp_dir)
    logger('This is a test for log_plot')    
    
    logger.log_single('var1',var1)

def test_log_plot():
    var1 = np.random.rand(100)
    var2 = 3 + np.random.rand(100)
    var3 = 6 + np.random.rand(100)
    
    logger = lognflow(temp_dir)
    logger('This is a test for log_plot')    
    
    logger.log_plot(parameter_name = 'var1', 
                    parameter_value_list = var1)
    
    logger.log_plot(parameter_name = 'vars', 
                    parameter_value_list = [var1, var2, var3])
    
def test_log_hist():
    var1 = np.random.rand(10000)
    var2 = 3 + np.random.rand(10000)
    var3 = 6 + np.random.rand(10000)
    
    logger = lognflow(temp_dir)
    logger('This is a test for log_hist')    
    
    logger.log_hist(parameter_name = 'var1', 
                    parameter_value_list = var1,
                    n_bins = 100)
    
    logger.log_hist(parameter_name = 'vars', 
                    parameter_value_list = [var1, var2, var3],
                    n_bins = 100)
    
def test_log_scatter3():
    var1 = np.random.rand(100)
    var2 = 3 + np.random.rand(100)
    var3 = 6 + np.random.rand(100)

    var3d = np.array([var1, var2, var3])
    
    logger = lognflow(temp_dir)
    logger('This is a test for log_scatter3')    
    
    logger.log_scatter3('var3d', var3d)    
    
def test_log_plt():
    plt.imshow(np.random.rand(100, 100))
    logger = lognflow(temp_dir)
    logger('This is a test for log_plt')    
    logger.log_plt('var3d')        
    
def test_log_hexbin():
    var1 = np.random.randn(10000)
    var2 = 3 + np.random.randn(10000)

    logger = lognflow(temp_dir)
    logger('This is a test for log_hexbin')    
    
    logger.log_hexbin('hexbin', [var1, var2])    

def test_log_imshow():
    logger = lognflow(temp_dir)
    logger('This is a test for log_imshow')    
    logger.log_imshow('var3d', np.random.rand(100, 100))    

def test_prepare_stack_of_images():
    stack_1 = np.random.rand(8, 100, 100, 9)
    stack_1[:, :1, :, :] = np.nan
    stack_1[:, :, :1, :] = np.nan
    stack_1[:, -1:, :, :] = np.nan
    stack_1[:, :, -1:, :] = np.nan
    stack_2 = np.random.rand(8, 100, 100)
    stack_3 = np.random.rand(8, 100, 100, 3, 9)
    stack_3[:, :1] = np.nan
    stack_3[:, :, :1] = np.nan
    stack_3[:, -1:] = np.nan
    stack_3[:, :, -1:] = np.nan
    
    list_of_stacks123 = [stack_1, stack_2, stack_3]
    logger = lognflow(temp_dir)
    logger('This is a test for prepare_stack_of_images')
    list_of_stacks123 = logger.prepare_stack_of_images(list_of_stacks123)
    logger.log_canvas('canvas_before_handling', list_of_stacks123)

def test_log_canvas():
    imgs=[]
    for _ in range(5):
        _imgs = np.random.rand(5, 100, 100)
        _imgs[:, 50, 50] = 2
        imgs.append(_imgs)
    
    logger = lognflow(temp_dir)
    logger('This is a test for log_canvas')    
    logger(f'imgs.shape: {imgs[0].shape}')

    logger.log_single('test_param1', _imgs)
    logger.log_single('test_param2/', _imgs)
    logger.log_single('test_param3\\', _imgs)
    logger.log_single('test_param4\d', _imgs)
    logger.log_single('test_param4\d2\\', _imgs)
    logger.log_single('test_param4\d2/', _imgs)
    logger.log_single('test_param4\d2/e', _imgs)

    logger.log_canvas(parameter_name = 'test_canvas\\', 
                      list_of_stacks = imgs, 
                      text_as_colorbar = True)
    
def test_log_confusion_matrix():
    from sklearn.metrics import confusion_matrix
    
    vec1 = np.random.rand(10000) > 0.8
    vec2 = np.random.rand(10000) > 0.2
    
    cm = confusion_matrix(vec1, vec2, normalize='all')
    logger = lognflow(temp_dir)
    logger('This is a test for log_confusion_matrix')
    logger.log_confusion_matrix('cm1', cm, title = 'test_log_confusion_matrix')

def test_rename():
    logger = lognflow(temp_dir)
    logger('This is a test for test_rename')
    logger.rename(logger.log_dir.name + '_new_name')
    logger('This is another test for test_rename')
    
if __name__ == '__main__':
    test_lognflow()
    test_log_var()
    test_log_animation()
    test_log_single()
    test_log_plot()
    test_log_hist()
    test_log_scatter3()
    test_log_plt()
    test_log_hexbin()
    test_log_imshow()
    test_prepare_stack_of_images()
    test_log_canvas()
    test_log_confusion_matrix()
    test_rename()


#!/usr/bin/env python

"""Tests for `lognflow` package."""

import pytest

from lognflow import lognflow, select_directory, logviewer, printprogress

import numpy as np

@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')

def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string

def test_printprogress():
    N = 10000000
    pprog = printprogress(N)
    for _ in range(N):
        pprog()
    # assert input('Did it show you a progress bar? (y for yes)')=='y'
    
if __name__ == '__main__':
    test_printprogress()
    
    



#!/usr/bin/env python

"""Tests for `lognflow` package."""

import pytest

from lognflow import lognflow, select_directory, logviewer, printprogress

import numpy as np

@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')

def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string

def test_logviewer():
    temp_dir = select_directory()
    logger = lognflow(temp_dir)
    logger('Well this is a test for logviewer')
    
    log_dir =  select_directory(temp_dir)
    logged = logviewer(log_dir, logger)
    print(logged.get_variable('test_param'))
    print(logged.get_log_text())
    
if __name__ == '__main__':
    test_logviewer()
    
    



import pathlib
import time
# import logging
import numpy as np
import itertools
from os import sep as os_sep

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib import animation

from collections import namedtuple

varinlog = namedtuple('varinlog',
                      ['data_array', 
                      'time_array', 
                      'cur_index', 
                      'filesplit_cnt', 
                      'save_as',
                      'log_counter_limit'])

class lognflow:
    """ lognflow logs your workflow
    
    Indeed we need a simple lognflow that works on local machine disk.
    This is a minimalist approach to solve this problem. 
    """
    
    def __init__(self, 
                 logs_root : pathlib.Path = None,
                 log_dir : pathlib.Path = None,
                 print_text = True,
                 main_log_name = 'main_log'):
        """ lognflow construction
        
        The lognflow is an easy way to log your variables on a local disk..
        It will also count files up when they get too large. It also can save
        np.arrays in npz format which is better than csv or what not.
        
        Parameters
        ----------
            logs_root: pathlib.Path
                This is the root directory for all logs. We will use the time.time()
                to create a log directory for each instance of the lognflow. 
            log_dir: pathlib.Path
                This is the final directory path for the log files. 
            
            note:
            ^^^^^^^^
                One of the variables logs_root or log_dir must be provided.

            print_text ; bool
                If True, everything that is logged as text will be printed as well
        """
        self._init_time = time.time()

        if(log_dir is None):
            if(logs_root is None):
                logs_root = 'logs'
            self.log_dir = pathlib.Path(logs_root) / f'{int(self._init_time):d}/'
        else:
            self.log_dir = pathlib.Path(log_dir)
        if(not self.log_dir.is_dir()):
            self.log_dir.mkdir(parents = True, exist_ok = True)
        if(not self.log_dir.is_dir()):
            return
        
        self._print_text = print_text
        self._loggers_dict = {}
        self._vars_dict = {}
        self._single_var_call_cnt = 0

        self.log_name = main_log_name
    
    def rename(self, new_dir:str):
        for log_name in self._loggers_dict.keys():
            _logger, _, _, _ = self._loggers_dict[log_name]
            _logger.close()
        self.log_dir = self.log_dir.rename(self.log_dir.parent / new_dir)
        for log_name in self._loggers_dict.keys():
            _logger, n_lines_limit, log_text_cnt, log_file_id = \
                self._loggers_dict[log_name]
            log_fpath = self.log_dir / log_file_id
            _logger = open(log_fpath, 'a')
            self._loggers_dict[log_name] = [_logger, 
                                       n_lines_limit, 
                                       log_text_cnt, 
                                       log_file_id]
    
    def _log_text_handler(self, log_name = None, 
                         title = None, 
                         n_lines_limit: int = int(1e+7),
                         time_in_file_name = True):

        if(log_name in self._loggers_dict.keys()):
            _logger, n_lines_limit, log_text_cnt, log_file_id = \
                self._loggers_dict[log_name]
            _logger.close()
        
        log_text_cnt = 0
        if(time_in_file_name):
            log_file_id = log_name + f'_{int(time.time()):d}.txt'
        else:
            log_file_id = log_name + '.txt'
        log_fpath = self.log_dir / log_file_id
        _logger = open(log_fpath, 'w')
        if(title is not None):
            _logger.write(title + '\n')
        self._loggers_dict[log_name] = [_logger, 
                                       n_lines_limit, 
                                       log_text_cnt, 
                                       log_file_id]
        
    def log_text(self, 
                 to_be_logged : str = '\n', 
                 log_name : str = None,
                 log_time_stamp = True,
                 print_text = None,
                 title = None, 
                 n_lines_limit: int = int(1e+7),
                 time_in_file_name = True,
                 **_):
        time_time = time.time() - self._init_time

        if(log_name is None):
            log_name = self.log_name
        
        if(print_text is None):
            print_text = self._print_text
        
        if(print_text):
            if(log_time_stamp):
                print(f'T:{time_time:>6.6f}|')
            print(f' log name: {log_name}')
            print(to_be_logged)
        
        if not (log_name in self._loggers_dict.keys()):
            self._log_text_handler(log_name, 
                                   title = title, 
                                   n_lines_limit = n_lines_limit,
                                   time_in_file_name = time_in_file_name)

        _logger, n_lines_limit, log_text_cnt, log_file_id = \
            self._loggers_dict[log_name]
        if(log_time_stamp):
            _logger.write(f'T:{time_time:>6.6f}| ')
        if isinstance(to_be_logged, np.ndarray):
            try:
                _logger.write('numpy.ndarray')
                if(to_be_logged.size()>100):
                    _logger.write(', The first and last 50 elements:\n')
                    to_be_logged = to_be_logged.ravel()
                    _logger.write(np.array2string(to_be_logged[:50]))
                    _logger.write(' ... ')
                    _logger.write(np.array2string(to_be_logged[-50:]))
                else:
                    _logger.write(':\n' + np.array2string(to_be_logged))
            except:
                _logger.write(' not possible to log ' + log_name + '\n')
        else:
            _logger.write(to_be_logged + '\n')
        log_text_cnt += len(to_be_logged)
        self._loggers_dict[log_name] = [_logger, 
                                       n_lines_limit, 
                                       log_text_cnt, 
                                       log_file_id]

        if(log_text_cnt >= n_lines_limit):
            self._log_text_handler(log_name)
            _, _, _, log_file_id = self._loggers_dict[log_name]
        return log_file_id
    
    def __call__(self, *args, **kwargs):
        self.log_text(*args, **kwargs)
                    
    def _prepare_param_dir(self, parameter_name):
        
        try:
            _ = parameter_name.split()
        except:
            self.log_text('The parameter name is not a string')
            self.log_text(f'Its type is {type(parameter_name)}')
            self.log_text(f'It is {parameter_name}')
        assert len(parameter_name.split()) == 1, \
            self.log_text(\
                  f'The variable name {parameter_name} you chose is splitable' \
                + f' I can split it into {parameter_name.split()}'             \
                + ' Make sure you dont use space, tab, or ....'                \
                + ' If you are using single backslash, e.g. for windows'       \
                + ' folders, replace it with \\ or make it a literal string'   \
                + ' by putting an r before the variable name.')
        
        is_dir = (parameter_name[-1] == '/') | (parameter_name[-1] == '\\') \
                 | (parameter_name[-1] == r'/') | (parameter_name[-1] == r'\\')
        param_dir = self.log_dir /  parameter_name
        
        if(is_dir):
            param_name = ''
        else:
            param_name = param_dir.name
            param_dir = param_dir.parent
        if(not param_dir.is_dir()):
            self.log_text(f'Creating directory: {param_dir.absolute()}')
            param_dir.mkdir(parents = True, exist_ok = True)
        return(param_dir, param_name)                    

    def _get_log_counter_limit(self, param, log_size_limit):
        cnt_limit = int(log_size_limit/(param.size*param.itemsize))
        return cnt_limit

    def log_var(self, parameter_name : str, parameter_value, 
                save_as='npz', log_size_limit: int = int(1e+7)):
        time_time = time.time() - self._init_time
        
        try:
            _ = parameter_value.shape
        except:
            parameter_value = np.array([parameter_value])
        
        log_counter_limit = self._get_log_counter_limit(\
            parameter_value, log_size_limit)

        if(parameter_name in self._vars_dict.keys()):
            _var = self._vars_dict[parameter_name]
            data_array, time_array, cur_index, \
                filesplit_cnt, save_as, log_counter_limit = _var
            cur_index += 1
        else:
            filesplit_cnt = time.time()
            cur_index = 0

        if(cur_index >= log_counter_limit):
            self._log_var_save(parameter_name)
            filesplit_cnt = time.time()
            cur_index = 0

        if(cur_index == 0):
            data_array = np.zeros((log_counter_limit, ) + parameter_value.shape,
                                  dtype = parameter_value.dtype)
            time_array = np.zeros(log_counter_limit)
        
        try:
            time_array[cur_index] = time_time
        except:
            self.log_text(f'current index {cur_index} cannot be used in the logger')
        if(parameter_value.shape == data_array[cur_index].shape):
            data_array[cur_index] = parameter_value
        else:
            self.log_text(\
                f'Shape of variable {parameter_name} cannot change '\
                f'from {data_array[cur_index].shape} '\
                f'to {parameter_value.shape}. Coppying from the last time.')
            data_array[cur_index] = data_array[cur_index - 1]
        self._vars_dict[parameter_name] = varinlog(data_array, 
                                                  time_array, 
                                                  cur_index,
                                                  filesplit_cnt,
                                                  save_as,
                                                  log_counter_limit)

    def _log_var_save(self, parameter_name : str):
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        
        _var = self._vars_dict[parameter_name]
        if(_var.save_as == 'npz'):
            fpath = param_dir / (f'{param_name}_{_var.filesplit_cnt}.npz')
            np.savez(fpath,
                time_array = _var.time_array,
                data_array = _var.data_array)
        elif(_var.save_as == 'txt'):
            fpath = param_dir / (f'{param_name}_time_{_var.filesplit_cnt}.txt')
            np.savetxt(fpath, _var.data_array)
            fpath = param_dir / (f'{param_name}_data_{_var.filesplit_cnt}.txt')
            np.savetxt(fpath, _var.data_array)
        return fpath
    
    def log_var_flush(self):
        for parameter_name in self._vars_dict.keys():
            self._log_var_save(parameter_name)
    
    def log_animation(self, parameter_name : str, stack, 
                         interval=50, blit=False, 
                         repeat_delay = None, dpi=100):
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
                    
        fpath = param_dir / (fname +'.gif')
        fig, ax = plt.subplots()
        ims = []
        for img in stack:    
            im = ax.imshow(img, animated=True)
            plt.xticks([]),plt.yticks([])
            ims.append([im])
        ani = animation.ArtistAnimation(\
            fig, ims, interval = interval, blit = blit,
            repeat_delay = repeat_delay)    
        ani.save(fpath, dpi = dpi, 
                 writer = animation.PillowWriter(fps=int(1000/interval)))
        
        return fpath

    def log_single(self, parameter_name : str, 
                         parameter_value,
                         save_as = 'npy'):
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(save_as == 'mat'):
            if(len(param_name) == 0):
                param_name = 'data'
        
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
            
        if(save_as == 'npy'):
            fpath = param_dir / (fname + '.npy')
            np.save(fpath, parameter_value)
        elif(save_as == 'txt'):
            fpath = param_dir / (fname + '.txt')
            np.savetxt(fpath, parameter_value)
        elif(save_as == 'mat'):
            fpath = param_dir / (fname + '.mat')
            from scipy.io import savemat
            savemat(fpath, {f'{param_name}' :parameter_value})
        elif('torch'):
            fpath = param_dir / (fname + '.torch')
            from torch import save as torch_save
            torch_save(parameter_value.state_dict(), fpath)
        return fpath
    
    def log_plot(self, parameter_name : str, 
                       parameter_value_list,
                       x_value = None,
                       image_format='jpeg', dpi=1200):
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
            
        fpath = param_dir / (fname + '.jpg')
        
        try:
            if not isinstance(parameter_value_list, list):
                parameter_value_list = [parameter_value_list]
                
            if(x_value is not None):
                if not isinstance(x_value, list):
                    x_value = [x_value]
            
            if( (len(x_value) != len(parameter_value)) & \
                (len(x_value) != 1)):
                self.log_text(f'x_value for {parameter_name} should have ' \
                    + 'length of 1 or the same as parameters list.')
                raise ValueError
            
            for list_cnt, parameter_value in enumerate(parameter_value_list):
                if(x_value is None):
                    plt.plot(parameter_value)
                else:
                    if(len(x_value) == len(parameter_value)):
                        plt.plot(x_value[list_cnt], parameter_value)
                    else:
                        plt.plot(x_value[0], parameter_value)
                        
            plt.savefig(fpath, format=image_format, dpi=dpi)
            plt.close()
            return fpath
        except:
            self.log_text(f'Cannot plot variable {parameter_name}.')
            return None
    
    def log_hist(self, parameter_name : str, 
                       parameter_value_list,
                       n_bins = 10,
                       image_format='jpeg', dpi=1200):
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
            
        fpath = param_dir / (fname + '.jpg')
        
        try:
            if not isinstance(parameter_value_list, list):
                parameter_value_list = [parameter_value_list]
                
            for list_cnt, parameter_value in enumerate(parameter_value_list):
                bins, edges = np.histogram(parameter_value, n_bins)
                plt.bar(edges[:-1], bins, width =np.diff(edges).mean(), alpha=0.5)
                plt.plot(edges[:-1], bins)
                        
            plt.savefig(fpath, format=image_format, dpi=dpi)
            plt.close()
            return fpath
        except:
            self.log_text(f'Cannot plot variable {parameter_name}.')
            return None
    
    def log_scatter3(self, parameter_name : str, 
                       parameter_value,
                       image_format='jpeg', dpi=1200):
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
            
        fpath = param_dir / (fname + '.jpg')
        
        try:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.scatter(parameter_value[0], parameter_value[1], parameter_value[2])
            plt.savefig(fpath, format=image_format, dpi=dpi)
            plt.close()
            return fpath
        except:
            self.log_text(f'Cannot plot variable {parameter_name}.')
            return None
    
    def log_plt(self, 
                parameter_name : str, 
                image_format='jpeg', dpi=1200):
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
            
        fpath = param_dir / (fname + '.jpg')
        
        try:
            plt.savefig(fpath, format=image_format, dpi=dpi)
            plt.close()
            return fpath
        except:
            self.log_text(f'Cannot save the plt instance {parameter_name}.')
            return None
    
    def log_hexbin(self, parameter_name : str, parameter_value,
                   gridsize = 20, image_format='jpeg', dpi=1200):
        
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
        fpath = param_dir / (f'{fname}.{image_format}')
        
        plt.figure()
        plt.hexbin(parameter_value[0], 
                   parameter_value[1], 
                   gridsize = gridsize)
        plt.savefig(fpath, format=image_format, dpi=dpi)
        plt.close()    
        return fpath
    
    def log_imshow(self, parameter_name : str, parameter_value,
                   image_format='jpeg', dpi=1200):
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
            
        fpath = param_dir / (f'{fname}.{image_format}')
        
        parameter_value = np.squeeze(parameter_value)
        parameter_value_shape = parameter_value.shape
        n_dims = len(parameter_value_shape)
        
        FLAG_img_ready = False
        if(n_dims == 2):
            FLAG_img_ready = True
        elif(n_dims == 3):
            if(parameter_value_shape[2] == 3):
                FLAG_img_ready = True
        elif(n_dims == 4):
            parameter_value = parameter_value.swapaxes(1,2)
            new_shape = parameter_value.shape
            parameter_value = parameter_value.reshape(new_shape[0] * new_shape[1],
                                                      new_shape[2] * new_shape[3])
            FLAG_img_ready = True
        elif(n_dims == 5):
            if(parameter_value_shape[4] == 3):
                parameter_value = parameter_value.swapaxes(1,2)
                new_shape = parameter_value.shape
                parameter_value = parameter_value.reshape(\
                    new_shape[0] * new_shape[1],
                    new_shape[2] * new_shape[3],
                    new_shape[4])
                FLAG_img_ready = True
        
        if(FLAG_img_ready):
            plt.imshow(parameter_value)
            plt.colorbar()
            plt.savefig(fpath, format = image_format, dpi=dpi)
            plt.close()
            return fpath
        else:
            self.log_text(f'Cannot plot variable {parameter_name} with shape' + \
                          f'{parameter_value.shape}')
            return

    def _handle_images_stack(self, stack, nan_borders):
        if(len(stack.shape) == 2):
            canv = np.expand_dims(stack, axis=0)
        elif(len(stack.shape) == 3):
            canv = stack
        elif((len(stack.shape) == 4) | (len(stack.shape) == 5)):
            if(len(stack.shape) == 4):
                n_imgs, n_R, n_C, n_ch = stack.shape
                if(n_ch == 3):
                    canv = stack
            if(len(stack.shape) == 5):
                n_imgs, n_R, n_C, is_rgb, n_ch = stack.shape
                if(is_rgb != 3):
                    return None
            square_side = int(np.ceil(np.sqrt(n_ch)))
            new_n_R = n_R * square_side
            new_n_C = n_C * square_side
            if(len(stack.shape) == 4):
                canv = np.zeros((n_imgs, new_n_R, new_n_C), dtype = stack.dtype)
            if(len(stack.shape) == 5):
                canv = np.zeros((n_imgs, new_n_R, new_n_C, 3), dtype = stack.dtype)
            used_ch_cnt = 0

            stack[:, :1] = np.nan
            stack[:, :, :1] = np.nan
            stack[:, -1:] = np.nan
            stack[:, :, -1:] = np.nan
            
            for rcnt in range(square_side):
                for ccnt in range(square_side):
                    ch_cnt = rcnt + square_side*ccnt
                    if (ch_cnt<n_ch):
                        canv[:, rcnt*n_R : (rcnt + 1)*n_R,
                                ccnt*n_C : (ccnt + 1)*n_C] = \
                            stack[..., used_ch_cnt]
                        used_ch_cnt += 1
        else:
            return None
        return canv

    def prepare_stack_of_images(self, list_of_stacks, nan_borders = True):        
        for cnt, stack in enumerate(list_of_stacks):
            stack = self._handle_images_stack(stack, nan_borders = nan_borders)
            if(stack is None):
                return
            list_of_stacks[cnt] = stack
        return(list_of_stacks)

    def log_canvas(self, 
                   parameter_name : str,
                   list_of_stacks : list,
                   list_of_masks = None,
                   figsize_ratio = 1,
                   text_as_colorbar = False,
                   use_colorbar = False,
                   image_format='jpeg', 
                   dpi=600):
        
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
        fpath = param_dir / (f'{fname}.{image_format}')
                
        try:
            _ = list_of_stacks.shape
            list_of_stacks = [list_of_stacks]
        except:
            pass
        n_stacks = len(list_of_stacks)
        if(list_of_masks is not None):
            n_masks = len(list_of_masks)
            assert (n_masks == n_stacks), \
                f'the number of masks, {n_masks} and ' \
                + f'stacks {n_stacks} should be the same'
        
        n_imgs = list_of_stacks[0].shape[0]
                
        plt.figure(figsize = (n_imgs*figsize_ratio,n_stacks*figsize_ratio))
        gs1 = gridspec.GridSpec(n_stacks, n_imgs)
        if(use_colorbar):
            gs1.update(wspace=0.25, hspace=0)
        else:
            gs1.update(wspace=0.025, hspace=0) 
        
        canvas_mask_warning = False
        for img_cnt in range(n_imgs):
            for stack_cnt in range(n_stacks):
                ax1 = plt.subplot(gs1[stack_cnt, img_cnt])
                plt.axis('on')
                ax1.set_xticklabels([])
                ax1.set_yticklabels([])
                data_canvas = list_of_stacks[stack_cnt][img_cnt].copy()
                vmin = data_canvas.min()
                vmax = data_canvas.max()
                if(list_of_masks is not None):
                    mask = list_of_masks[stack_cnt]
                    if(mask is not None):
                        if(data_canvas.shape == mask.shape):
                            data_canvas[mask==0] = 0
                            vmin = data_canvas[mask>0].min()
                            vmax = data_canvas[mask>0].max()
                        elif(not canvas_mask_warning):
                            self.log_text(\
                                'The mask shape is different from the canvas.' \
                                + ' No mask will be applied.')
                            canvas_mask_warning = True
                im = ax1.imshow(data_canvas, vmin = vmin, vmax = vmax)
                if(text_as_colorbar):
                    ax1.text(data_canvas.shape[0]*0,
                             data_canvas.shape[1]*0.05,
                             f'{data_canvas.max():.6f}', 
                             color = 'yellow',
                             fontsize = 2)
                    ax1.text(data_canvas.shape[0]*0,
                             data_canvas.shape[1]*0.5, 
                             f'{data_canvas.mean():.6f}', 
                             color = 'yellow',
                             fontsize = 2)
                    ax1.text(data_canvas.shape[0]*0,
                             data_canvas.shape[1]*0.95, 
                             f'{data_canvas.min():.6f}', 
                             color = 'yellow',
                             fontsize = 2)
                if(use_colorbar):
                    cbar = plt.colorbar(im, ax=ax1, fraction=0.046, pad=0.04)
                    cbar.ax.tick_params(labelsize=1)
                ax1.set_aspect('equal')
        
        plt.savefig(fpath, format=image_format, dpi=dpi)
        plt.close()
        return fpath

    def log_confusion_matrix(self,
                             parameter_name : str,
                             cm,
                             target_names = None,
                             title='Confusion matrix',
                             cmap=None,
                             figsize = None,
                             image_format = 'jpeg',
                             dpi = 1200):
        """
        given a sklearn confusion matrix (cm), make a nice plot
    
        Arguments
        ---------
        cm:           confusion matrix from sklearn.metrics.confusion_matrix
    
        target_names: given classification classes such as [0, 1, 2]
                      the class names, for example: ['high', 'medium', 'low']
    
        title:        the text to display at the top of the matrix
    
        cmap:         the gradient of the values displayed from matplotlib.pyplot.cm
                      (http://matplotlib.org/examples/color/colormaps_reference.html)
                      plt.get_cmap('jet') or plt.cm.Blues
    
        Usage
        -----
        plot_confusion_matrix(\
            cm           = cm,                  # confusion matrix created by
                                                # sklearn.metrics.confusion_matrix
            target_names = y_labels_vals,       # list of names of the classes
            title        = best_estimator_name) # title of graph
    
        Citiation
        ---------
        http://scikit-learn.org/stable/auto_examples/model_selection/
                                                           plot_confusion_matrix.html
    
        """
        time_time = time.time() - self._init_time
        param_dir, param_name = self._prepare_param_dir(parameter_name)
        if(len(param_name) > 0):
            fname = f'{param_name}_{time_time}'
        else:
            fname = f'{time_time}'
        fpath = param_dir / (f'{fname}.{image_format}')
        
        accuracy = np.trace(cm) / np.sum(cm).astype('float')
        misclass = 1 - accuracy
    
        if figsize is None:
            figsize = np.ceil(cm.shape[0]/3)
    
        if target_names is None:
            target_names = [chr(x + 65) for x in range(cm.shape[0])]
    
        if cmap is None:
            cmap = plt.get_cmap('Blues')
    
        plt.figure(figsize=(4*figsize, 4*figsize))
        im = plt.imshow(cm, interpolation='nearest', cmap=cmap)
    
        if target_names is not None:
            tick_marks = np.arange(len(target_names))
            plt.xticks(tick_marks, target_names, rotation=45)
            plt.yticks(tick_marks, target_names)
    
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            clr = np.array([1, 1, 1, 0]) \
                  * (cm[i, j] - cm.min()) \
                      / (cm.max() - cm.min()) + np.array([0, 0, 0, 1])
            plt.text(j, i, f"{cm[i, j]:2.02f}", horizontalalignment="center",
                     color=clr)
        
        plt.ylabel('True label')
        plt.xlabel('Predicted label\naccuracy={:0.4f}; ' \
                   + 'misclass={:0.4f}'.format(accuracy, misclass))
        plt.title(title)
        plt.colorbar(im,fraction=0.046, pad=0.04)
        plt.tight_layout()
        plt.savefig(fpath, format=image_format, dpi=dpi)
        plt.close()
        return fpath
    
    def __del__(self):
        self.log_var_flush()
        
def select_directory(start_directory = './'):
    from PyQt5.QtWidgets import QFileDialog, QApplication
    _ = QApplication([])
    log_dir = QFileDialog.getExistingDirectory(
        None, "Open a folder", start_directory, QFileDialog.ShowDirsOnly)
    return(log_dir)


from math import ceil
from time import time

class printprogress:
    """
    While there are modules that use \r to show a progress bar, 
    there are cases e.g. a print_function or an ssh terminals that do not 
    support \r. In such cases, if something is written at the end of 
    the line, it cannot be deleted. The following code provides a good 
    looking simple title for the progress bar and shows the limits and 
    is very simple to use. Define the object with number of steps of the loop 
    and then  call it in every iteration of the loop.
    """
    def __init__(self, 
                 n_steps, 
                 numTicks = 78,
                 title = None,
                 print_function = print):
        """
            n_steps: int
                usually the number of iterations in the for loop
            numTicks: int
                The number of charachters in a row of the screen - 2
                default: 78
            title : str 
                The title of progress bar.
                default: f'Progress bar for {n_steps} steps in {numTicks} ticks'
            print_function:
                print_function must be callable with a string and should not add
                \n to the its input just because its been called.
        """
        
        self.in_print_function = print_function
        
        if(n_steps != int(n_steps)):
            self._print_function('textProgBar takes integers no less than 2 ')
            n_steps = int(n_steps)
        if(n_steps<2):
            n_steps = 2
        if (title is None):
            title = f'Progress bar for {n_steps} steps in {numTicks} ticks'
        self.FLAG_ended = False
        self.FLAG_warning = False
        self.startTime = time()
        self.ck = 0
        self.prog = 0
        self.n_steps = n_steps
        if(numTicks < len(title) + 2 ):
            self.numTicks = len(title)+2
        else:
            self.numTicks = numTicks
        for _ in range(self.numTicks):
            self._print_function('_', end='')
        self._print_function(' ')
        self._print_function('/', end='')
        for idx in range(self.numTicks - len(title)):
            if(idx==int((self.numTicks - len(title))/2)):
                self._print_function(title, end='')
            else:
                self._print_function(' ', end='')
        self._print_function(' \\')
        self._print_function(' ', end = '')
    
    def _print_function(self, text, end='\n'):
        if (self.in_print_function == print):
            print(text, end = end, flush = True)
        else:
            self.in_print_function(text)
        
    def __call__(self, ck=1):
        """ ticking the progress bar
            just call the object and the progress bar moves one agead.
                   p                   x      
            |-----------|--------------------------|
                       ck                       n_steps
            As such  x/(n-c) = p/c => x = p(n/c - 1)
        """
        if(self.FLAG_ended):
            if(not self.FLAG_warning):
                self.FLAG_warning = True
                self._print_function('-' * self.numTicks)
        else:
            self.ck += ck
            if(self.ck >= self.n_steps):
                self.end()
            else:
                cProg = int(self.numTicks*self.ck/(self.n_steps-1)/3)    
                #3: because 3 charachters are used
                while((self.prog < cProg) & (not self.FLAG_ended)):
                    self.prog += 1
                    passedTime = time() - self.startTime
                    remTimeS = passedTime * ( self.n_steps / self.ck - 1)
                    if(remTimeS>=5940):
                        progStr = "%02d" % int(ceil(remTimeS/3600))
                        self._print_function(progStr, end='')
                        self._print_function('h', end='')
                    elif(remTimeS>=99):
                        progStr = "%02d" % int(ceil(remTimeS/60))
                        self._print_function(progStr, end='')
                        self._print_function('m', end='')
                    elif(remTimeS>0):
                        progStr = "%02d" % int(ceil(remTimeS))
                        self._print_function(progStr, end='')
                        self._print_function('s', end='')
                    else:
                        self.end()
    def end(self):
        if(not self.FLAG_ended):
            self._print_function('')
            self.FLAG_ended = True


"""Top-level package for lognflow."""

__author__ = """Alireza Sadri"""
__email__ = 'arsadri@gmail.com'
__version__ = '0.1.0'

from .lognflow import lognflow, select_directory
from .logviewer import logviewer
from .printprogress import printprogress